import numpy as np

import tools as T

# %%
__all__ = [
'equal',
'merge_dict',
'Squeezer'
]

# %%
def equal(array, axis=None):
    """
    Check if all elements in the array are equal along the specified axis.

    Parameters
    ----------
    array : array-like
        Input array to check for equality.
    axis : int, optional
        Axis along which to check for equality. If None, check the entire array.

    Returns
    -------
    bool or ndarray of bool
        if axis is None, return True if all elements are equal, False otherwise.
        if axis is specified, return an array of bools with axis dimension collapsed 
    """
    array = np.asarray(array)
    if axis is None:
        return np.all(array == array.flat[0])
    else:
        return np.all(array == np.expand_dims(array.take(0, axis=axis), axis=axis), axis=axis)

def merge_dict(ld):
    '''
    :param ld: list of dicts [{}, {}, {}]
    '''
    assert type(ld) == list, f'must give list of dicts, received: {type(ld)}'
    assert T.equal([list(d.keys()) for d in ld]), 'keys for every dict in the list of dicts must be the same'
    keys = ld[0].keys()
    merged_d = {}
    for k in keys:
        lv = [d[k] for d in ld] # list of values
        try:
            merged_d[k] = np.concatenate(lv, axis=0)
        except ValueError:
            merged_d[k] = np.array(lv)
    return merged_d

class Squeezer(object):
    """
    Warning: The class is very unstable, currently used as a temporary adjustment for concatenating / splitting batch dimension.

    Reshape numpy arrays or lists of numpy arrays.
    """
    def __init__(self):
        self.type = None
        self.ndim = None
        self.shape_original = None

    def squeeze(self, x):
        if isinstance(x, np.ndarray):
            self.type = np.ndarray
            self.ndim = x.ndim
            if self.ndim==2:
                self.shape_original = x.shape
            elif x.ndim==3:
                self.shape_original = x.shape
                x.shape = (self.shape_original[0]*self.shape_original[1], self.shape_original[2])
            else:
                raise Exception(f'x.ndim must be 2 or 3, received: {x.ndim}')
        elif isinstance(x, list):
            self.type = list
            if len(x) != 0:
                assert isinstance(x[0], np.ndarray), 'values in list must be np.ndarray'
                assert x[0].ndim==2, f'arrays in list must have ndim==2, received: {x[0].ndim}'
                self.ndim = 3
                self.shape_original = [x_.shape for x_ in x]
                x = np.concatenate(x, axis=0)
        else:
            raise TypeError('Input must be a numpy array or a list of numpy arrays.')
        
        return x

    def unsqueeze(self, x, strict=True):
        if self.type is None or self.ndim is None or self.shape_original is None:
            raise ValueError("Squeezer instance was not initialized correctly.")

        if self.ndim == 3:
            if self.type == np.ndarray:
                if strict:
                    if x.size == np.prod(self.shape_original):
                        x.shape = self.shape_original
                    else:
                        raise ValueError(f'Input shape {x.shape} cannot be reshaped into {self.shape_original}.')
                else:  # Allow flexibility in last dimension
                    expected_elements = np.prod(self.shape_original[:-1])
                    if x.size % expected_elements == 0:
                        x.shape = (*self.shape_original[:-1], -1)
                    else:
                        raise ValueError(f'Cannot reshape {x.shape} flexibly; incorrect number of elements.')
            elif self.type == list:
                split_indices = np.cumsum([shape[0] for shape in self.shape_original[:-1]])
                if x.shape[0] != sum(shape[0] for shape in self.shape_original):
                    raise ValueError(f'Input shape {x.shape} does not match expected concatenated shape.')
                x = np.split(x, split_indices, axis=0)

        return x

# %%
if __name__ == '__main__':
    ld = [{i:i+1 for i in range(5)} for j in range(3)]
    merge_dict(ld)
    ld = [{i:np.arange(i+1) for i in range(5)} for j in range(3)]
    merge_dict(ld)



from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import numpy as np

def standardize(x, scaler=None):
    """
    Standardize (z-score/gaussian normalization) x without having to make a StandardScaler() object.

    Parameters
    ----------
    x : ndarray of shape (n_observation, n_channel) 
        data to be standardized.
    scaler: sklearn.preprocessing.StandardScaler(), default=None
        scaler to use (that has already fit())

    Returns
    -------
    x_standardized : ndarray of shape (n_observation, n_channel)
        Normalized data
    scaler : sklearn.preprocessing.StandardScaler()
        Scaler used for normalization
    """
    if scaler is None:
        scaler = StandardScaler()
        x_standardized = scaler.fit_transform(x)
    else:
        x_standardized = scaler.transform(x)
    
    return x_standardized, scaler

def project_pca(x, var_threshold=None, n_pc=None, solver=None, random_state=0):
    """
    project x onto PC space without having to make a PCA() object.

    Parameters
    ----------
    x : ndarray of shape (n_observation, n_channel)
        data to be projected.
    solver: sklearn.decomposition.PCA(), default=None
        solver to use (that has already fit())

    Returns
    -------
    x_projected : ndarray of shape (n_observation, n_channel)
        Normalized data
    solver : sklearn.decomposition.PCA()
        Solver used for PCA projection
    """
    assert (var_threshold is None) ^ (n_pc is None), 'Only one of var_threshold or n_pc must be specified'

    # If svd_solver is not full, result is random
    if solver is None:
        solver = PCA(svd_solver='full', random_state=random_state)
        x_pc = solver.fit_transform(x)
    else:
        x_pc = solver.transform(x)

    if var_threshold is not None:
        n_pc = np.argmax(np.cumsum(solver.explained_variance_ratio_) > var_threshold) + 1
    elif n_pc is not None:
        pass
    x_pc = x_pc[:,:n_pc]

    return x_pc, solver

if __name__=='__main__':
    import tools as T

    x = np.random.rand(2,3,5)
    print(x.shape)
    squeezer = T.numpy.Squeezer()
    x = squeezer.squeeze(x)
    print(x.shape)
    x_s, scaler = T.sklearn.preprocessing.standardize(x)
    x_s = squeezer.unsqueeze(x_s)
    print(x_s.shape)

    # %%
    x = np.random.rand(2,3,5)
    print(x.shape)
    squeezer = T.numpy.Squeezer()
    x = squeezer.squeeze(x)
    print(x.shape)
    x_pc, solver = T.sklearn.preprocessing.project_pca(x, n_pc=3)
    x_pc = squeezer.unsqueeze(x_pc, strict=False)
    print(x_pc.shape)



import numpy as np
import torch
import tools as T


# TODO: change name mangic_cat() -> concat()
__all__ = [
'magic_cat',
]
# %%
def magic_cat(datas, axis=0, out=None):
    '''concatenate numpy or torch tensors'''
    assert T.equal([type(data) for data in datas]), 'Given data in datas must be the same type'
    datatype = type(datas[0])
    if datatype == torch.Tensor:
        return torch.cat(datas, dim=axis, out=out)
    elif datatype == np.ndarray or list:
        return np.concatenate(datas, axis=axis, out=out)
    else:
        raise Exception(f'Elements in data must be either [list, np.ndarray, torch.Tensor], received: {datatype}')


# Just use merge_dict and [value for value in variable]

# def dict_np_cat(d_array, axis=0):
#     '''
#     concatenates list of numpy array and returns concatenated dict of numpy arrays
#     '''
#     d_cat = {}
#     for key in d_array.keys():
#         d[key] = d
#     np.concatenate
#     return d_cat
#
# def dict_torchcat(d_array, axis=0):
#     pass
#
# def dict_magic_cat(d_array, axis=0):
#     pass



# import logging
import os
import warnings

import numpy as np
import matplotlib.colors as mcolors
from .. import numpy as tnp
# log = logging.getLogger(__name__)

__all__ = [
'get_xtick_seconds',
'multisave',
]

# %%
def get_xtick_seconds(ax, sr):
    '''
    :param ax: axis instance
    :param sec: second which maximum length indicates. it is assumed the xtick starts at 0
    '''
    # xticks = np.linspace(0,sec,len(ax_real.get_xticklabels())-2)
    # xticks = np.round(xticks, 3)
    # d = xticks[1]-xticks[0]
    # xticks = np.concatenate([[xticks[0] - d],xticks,[xticks[-1]+d]])
    # return xticks
    return ax.get_xticks() / sr

def multisave(fig, path, dpi=300):
    '''
    save given figure in .png, .eps, .svg format
    :param fig: pyplot figure instance
    :param path: path of the figure without extension (.extension)
    '''
    if '.' in os.path.basename(path):
        warnings.warn("Do not specify '.' in name. assuming it's part of the filename...")

        # log.warning('Do not specify extension in multisave(). Removing extension...')
        # path = os.path.splitext(path)[0]

    fig.savefig(path+'.png', dpi=dpi)
    fig.savefig(path+'.tiff', dpi=dpi)
    fig.savefig(path+'.eps')
    fig.savefig(path+'.svg')

def plot_values(valuetracker_list, ax=None):
    n_line = len(valuetracker_list)
    if ax is None:
        fig, ax = plt.subplots()
    color_list = it.cycle(mcolors.TABLEAU_COLORS)
    for valuetracker, color in zip(valuetracker_list, color_list):
        y_smooth = tnp.moving_mean(valuetracker.y, 9)
        ax.plot(valuetracker.x, valuetracker.y, color=color, alpha=0.4)
        ax.plot(valuetracker.x, y_smooth, color=color)
    return ax

def plot_trainval(valuetracker_list, ax=None):
    n_line = len(valuetracker_list)
    if ax is None:
        fig, ax = plt.subplots()
    color_list = ['tab:blue', 'tab:orange']
    line_list = ['-', 'x-']
    labels = ['train', 'val']
    for valuetracker, color, line, label in zip(valuetracker_list, color_list, line_list, labels):
        y_smooth = tnp.moving_mean(valuetracker.y, 9)
        ax.plot(valuetracker.x, valuetracker.y, color=color, alpha=0.4, label=label)
        ax.plot(valuetracker.x, y_smooth, line, color=color, label=label+'_smooth')
    return ax

if __name__ == '__main__':
    pass




import matplotlib.pyplot as plt

# %%
def imshow(tensor, **kwargs):
    '''
    tensor of (Channel, width, height)
    '''
    tensor=tensor.detach().cpu().numpy()
    fig, ax = plt.subplots()
    ax.imshow(tensor.transpose(1,2,0))

    return fig, ax



import random
import multiprocessing

import numpy as np
import torch

__all__ = [
'get_device',
'to',
'multiprocessing_device',
'nanparam',
'nangrad',
'param_same',
'print_rng_state',
'seed',
]

def get_device(network):
    return next(network.parameters()).device

# TODO: Merge into get_device
def device_of_optimizer(optimizer):
    return optimizer.param_groups[0]['params'][0].device

def to(data, *args, **kwargs):
    if isinstance(data, (list, tuple)):
        return tuple(getattr(tensor, 'to')(*args, **kwargs) for tensor in data)
    elif isinstance(data, dict):
        return {k: getattr(tensor, 'to')(*args, **kwargs) for k, tensor in data.items()}
    else:
        return data.to(*args, **kwargs)

def spread_device(gpu_id = None):
    '''
    what's this for?
    '''
    if gpu_id != None:
        gpu_id = gpu_id % torch.cuda.device_count()
    else:
        gpu_id = torch.cuda.default_stream().device.index
    device = torch.device(f"cuda:{gpu_id}" if torch.cuda.is_available() else "cpu")
    return device

def process_number():
    p=multiprocessing.current_process()
    try:
        worker_num = int(p.name.split('-')[-1]) # Fails if it's not integer
        return worker_num
    except ValueError:
        return -1

def multiprocessing_device(gpu_id = None):
    '''
    device setting for hydra multiprocessing
    '''
    # CPU
    if gpu_id == -1 or not torch.cuda.is_available():
        device = torch.device('cpu')
    # GPU
    else:
        if gpu_id == None:
            # use distributed device, or default device
            p=multiprocessing.current_process()
            #
            try:
                worker_num = int(p.name.split('-')[-1]) # Fails if it's not integer
                gpu_id = worker_num % torch.cuda.device_count()
                device = torch.device(f"cuda:{gpu_id}")
            except ValueError: # Parent process
                device = torch.cuda.default_stream().device
        else:
            device = torch.device(f'cuda:{gpu_id}')
        torch.cuda.set_device(device)
        torch.cuda.empty_cache()

    # log.info(f'device: {device}') # integrate logging?

    return device

def nanparam(model):
    for p in model.parameters():
        if torch.isnan(p).any():
            return True
    return False

def nangrad(model):
    for p in model.parameters():
        if torch.isnan(p.grad).any():
            return True
    return False

def param_same(model1, model2):
    for p1, p2 in zip(model1.parameters(), model2.parameters()):
        same = (p1==p2).all()
        if not same:
            return False
    return True

def print_rng_state(n=10, nonzero=True):
    rngs = torch.random.get_rng_state().numpy().tolist()
    if nonzero:
        rngs = [i for i in rngs if i != 0]
    print(rngs[:n])


def seed(random_seed, strict=False):
    '''

    '''
    np.random.seed(random_seed)
    torch.manual_seed(random_seed)

    if strict:
        # Following is verbose, but just in case.
        random.seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)

        # deterministic cnn
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False




__version__ = "2025.04.02.0"
from .tools import *
from . import os

'''
For third-party packages, import manually by something like:
import tools.pytorch
import tools.pytorch as tpytorch
import tools.pytorch as tpt
import tools.pytorch as Tpt
'''

# TODO: make animator in plot/




import matplotlib.pyplot as plt
import numpy as np
from . import numpy as tnp

class ValueTracker(object):
    """ ValueTracker."""

    def __repr__(self):
        return f'<ValueTracker>\nx: {self.x}\ny: {self.y}'

    def __init__(self):
        self.reset()

    def __len__(self):
        return len(self.y)

    def __iadd__(self, other):
        self.x.extend(other.x)
        self.y.extend(other.y)
        self.label.extend(other.label)
        self.n_step += len(other.x)
        return self

    def __add__(self, other):
        self = deepcopy(self)
        self.x.extend(other.x)
        self.y.extend(other.y)
        self.label.extend(other.label)
        self.n_step += len(other.x)
        return self

    def reset(self):
        self.x = []
        self.y = []
        self.label = []
        self.n_step = 0

    def numpy(self):
        return np.array(self.x), np.array(self.y), np.array(self.label)

    def step(self, x, y, label=None):
        if hasattr(x, '__len__'):
            assert hasattr(y, '__len__')
            assert len(x)==len(y)
            self.x.extend(x)
            self.y.extend(y)
            if label != None:
                assert len(y)==len(label)
                self.label.extend(label)
            self.n_step += len(x)

        else:
            self.x.append(x)
            self.y.append(y)
            if label != None:
                self.label.append(label)
            self.n_step += 1

    def plot(self, w=9, color='tab:blue', ax=None):
        x = np.array(self.x)
        y = np.array(self.y)
        y_smooth = tnp.moving_mean(y, w)
        if ax==None:
            ax = plt.gca()
        ax.plot(x, y, color=color, alpha=0.4)
        ax.plot(x, y_smooth, color=color)
        return ax

    def mean(self):
        return np.mean(self.y)
    def min(self):
        return np.min(self.y)
    def max(self):
        return np.max(self.y)

class AverageMeter(object):
    """Computes and stores the average and current value
    Variables
    ---------
    self.val
    self.avg
    self.sum
    self.count
    """
    # TODO: maybe keep track of each values? or just merge with valuetracker?

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def step(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

class DictList(object):
    """Dictionary of lists"""
    def __init__(self, keys):
        self._dict = {}

    def append(self, data):
        assert type(data) in [list, dict]
        if type(data)==dict:
            assert set(self._dict.keys())==set(data.keys()), f'allowed keys are: {self._dict.keys()}, received: {data.keys()}'
            for key, value in data.items():
                self._dict[key].append(item)
        else:
            warnings.warn('Appending with list is not recommended as this cannot ensure the data are being appended to the right place.')
            for key, value in zip(self._dict.items(), data):
                self._dict[key].append(value)



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as D

class ProxyDataset(D.Dataset):
    """
    ProxyDataset that does not copy the data of the original dataset.
    Useful for train/validation/test split without modifying the data
    (e.g. validation/test data is not preprocessed using train data)

    Parameters
    ----------
    dataset : torch.utils.data.Dataset object
        The original dataset
    idxs : list
        List of indices to use for the proxy dataset
    """
    def __init__(self, dataset, idxs):
        self.dataset = dataset
        self.idxs = idxs

        # Copy attributes from the original dataset
        keys = [key for key in dir(dataset) if key.startswith('__') is False]
        for key in keys:
            setattr(self, key, getattr(dataset, key))

    def __repr__(self):
        return f'ProxyDataset({self.dataset}, len: {len(self)}/{len(self.dataset)}({len(self)/len(self.dataset)*1e2:.0f}%))'

    def __getitem__(self, idx):
        return self.dataset[self.idxs[idx]]

    def __len__(self):
        return len(self.idxs)

def get_x_all(dataset):
    if hasattr(dataset, 'get_x_all'):
        return dataset.get_x_all()
    else:
        data_all = [data for data in dataset]
        if type(data_all[0]) == tuple: # More than 1 return value
            data_all = [data[0] for data in data_all]
        elif type(data_all[0]) == dict: # Dictionary return value
            data_all = [data['x'] for data in data_all]
        else: # 1 return value
            assert type(data_all[0]) == torch.Tensor

        return torch.stack(data_all, dim=0)

def get_y_all(dataset):
    if hasattr(dataset, 'get_y_all'):
        return dataset.get_y_all()
    else:
        data_all = [data for data in dataset]
        if type(data_all[0]) == tuple: # More than 1 return value
            data_all = [data[1] for data in data_all]
        elif type(data_all[0]) == dict: # Dictionary return value
            data_all = [data['y'] for data in data_all]
        else: # 1 return value
            raise Exception('No y data (2nd argument) found in the dataset')

        return torch.stack(data_all, dim=0)

def get_all(dataset):
    if hasattr(dataset, 'get_all'):
        return dataset.get_all()
    else:
        data_all = [data for data in dataset]
        if type(data_all[0]) == tuple: # More than 1 return value
            n_tuple = len(data_all[0])
            tensors = tuple([torch.stack([data[i] for data in data_all], dim=0) for i in range(n_tuple)])
        elif type(data_all[0]) == dict: # Dictionary return value
            keys = data_all[0].keys()
            tensors = {key: torch.stack([data[key] for data in data_all], dim=0) for key in keys}
        else: # 1 return value
            assert type(data_all[0]) == torch.Tensor
            tensors = torch.stack(data_all, dim=0)

    return tensors



from copy import deepcopy as dcopy
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

class FNN(nn.Module):
    '''
    basic FNN module

    Parameters
    ----------
    n_input: number of input (int)
        if n_input is None, then the LazyLinear() is set in the first layer.
    n_hidden_list: list of hidden neurons (list of int)
    activation_list: torch.nn activation function instances (nn object or list of nn object)

    Example
    -------
    >>> model = FNN(n_input=10, n_hidden_list=[8,6,5], activation_list=[nn.Sigmoid(), nn.ReLU(), nn.Tanh()])
    # n_hidden_list, activation_list corresponds to [h1, h2, output]
    >>> print(model)
    FNN(
      (fc): Sequential(
        (0): Linear(in_features=10, out_features=8, bias=True)
        (1): Sigmoid()
        (2): Linear(in_features=8, out_features=6, bias=True)
        (3): ReLU()
        (4): Linear(in_features=6, out_features=5, bias=True)
        (5): Tanh()
      )
    )
    '''
    def __init__(self, n_hidden_list, activation_list, n_input=None):
    # def __init__(self, n_input, n_hidden_list, activation_list):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'

        # 1st layer - Select Lazy if n_input is not specified
        if n_input is None:
            layers = [nn.Flatten(1), nn.LazyLinear(n_hidden_list[0]), activation_list[0]]
        else:
            layers = [nn.Flatten(1), nn.Linear(n_input, n_hidden_list[0]), activation_list[0]]
        # Hidden layers ~ Output layer
        for i in range(len(n_hidden_list) - 1):
            layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]), activation_list[i+1]])

        self.fc = nn.Sequential(*layers)

    def forward(self, x):
        '''x.shape()==(batch_size, feature_dim)'''
        return self.fc(x)

class FNN_Resnet(nn.Module):
    '''
    Resnet module

    Parameters
    ----------
    n_input : int
        number of input
    n_hidden_list : list of int
        list of hidden neurons
    activation : nn activation instance or list
        torch.nn activation function instances
    skip: int
        number of layers to skip. Interval for skipping
        ex) skip=2, then add skip layer every other 2 layers

    Example
    -------
    >>> model = Resnet(n_input=10, n_hidden_list=[8,6,5], activation=[nn.Sigmoid(), nn.ReLU(), nn.tanh()], skip=2)
    # activation corresponds to [h1, h2, output]
    >>> print(model)

    '''
    def __init__(self, n_hidden_list, activation_list, n_input=None, skip=2):
    # def __init__(self, n_input, n_hidden_list, activation_list, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert len(n_hidden_list) >= skip, 'number of layers must be equal or greater than skip. len(n_hidden_list): %s, skip: %s'%(len(n_hidden_list), skip)
        assert skip >= 2, 'skip needs to be: skip >= 2, given: %s'%(skip)

        # 1st layer
        if n_input is None:
            layers = [nn.LazyLinear(n_hidden_list[0])]
        else:
            layers = [nn.Linear(n_input, n_hidden_list[0])]
        # Hidden layers ~ Output layer
        layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]) for i in range(len(n_hidden_list)-1)])
        self.fc = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        # Bias already present at original layers
        self.skip = skip
        if n_input is None:
            n_skip_hidden_list = [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
            skip_layers = [nn.LazyLinear(n_skip_hidden_list[0], bias=False)] + [nn.Linear(n_skip_hidden_list[i], n_skip_hidden_list[i+1], bias=False) for i in range(len(n_skip_hidden_list)-1)]
        else:
            n_skip_hidden_list = [n_input] + [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
            skip_layers = [nn.Linear(n_skip_hidden_list[i], n_skip_hidden_list[i+1], bias=False) for i in range(len(n_skip_hidden_list)-1)]
        self.skip_layers = nn.ModuleList(skip_layers) # Need ModuleList, not list, so that the model can recognize this layers
        self.n_skip = len(n_hidden_list) // skip # number of skip layers
        # assert self.n_skip == len(self.skip_layers), 'something\'s wrong'

    def forward(self, x):
        x = x.flatten(1)
        last_x = x
        skip_layers = iter(self.skip_layers)
        for i, (layer, activation) in enumerate(zip(self.fc, self.activation_list)):
            if (i+1) % self.skip == 0:
                skip_layer = next(skip_layers)
                x = activation(layer(x)+skip_layer(last_x))
                last_x = x
            else:
                x = activation(layer(x))
        return x
    '''To do:
    - weight initialization? modify?
    - optimize using nn.Sequential() and itertools in forward. compute layers in group
    '''

class FNN_Densenet(nn.Module):
    def __init__(self, n_hidden_list, activation_list, n_input=None, skip=2):
    # def __init__(self, n_input, n_hidden_list, activation_list, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert len(n_hidden_list) >= skip, 'number of layers must be equal or greater than skip. len(n_hidden_list): %s, skip: %s'%(len(n_hidden_list), skip)
        assert skip >= 2, 'skip needs to be: skip >= 2, given: %s'%(skip)

        # 1st layer
        if n_input is None:
            layers = [nn.LazyLinear(n_hidden_list[0])]
        else:
            layers = [nn.Linear(n_input, n_hidden_list[0])]
        # Hidden layers ~ Output layer
        layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]) for i in range(len(n_hidden_list)-1)])
        self.fc = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        if n_input is None:
            n_skip_hidden_list = [None] + [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
            skip_layers = nn.ModuleDict()
            for i in range(1, len(n_skip_hidden_list)):
                for j in range(i):
                    if j==0:
                        skip_layers[str(j)+'_'+str(i)]=nn.LazyLinear(n_skip_hidden_list[i], bias=False)
                    else:
                        skip_layers[str(j)+'_'+str(i)]=nn.Linear(n_skip_hidden_list[j], n_skip_hidden_list[i], bias=False)
        else:
            n_skip_hidden_list = [n_input] + [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
            skip_layers = nn.ModuleDict()
            for i in range(1, len(n_skip_hidden_list)):
                for j in range(i):
                    skip_layers[str(j)+'_'+str(i)]=nn.Linear(n_skip_hidden_list[j], n_skip_hidden_list[i], bias=False)
        self.skip_layers = skip_layers

        n = len(n_skip_hidden_list)-1
        self.n_skip = int(n*(n+1)/2)
        assert self.n_skip == len(skip_layers), f'something wrong: n_skip({self.n_skip}), skip_layers({len(skip_layers)})'

    def forward(self, x):
        x_history = [x]
        dest = 1 # Destination index of skip layers
        for i, (layer, activation) in enumerate(zip(self.fc, self.activation_list)):
            if (i+1) % self.skip == 0:
                assert dest == len(x_history), f'Length of x_history{len(x_history)} and destination number{dest} must match'
                skip_x = torch.stack([self.skip_layers[str(src)+'_'+str(dest)](x_) for src, x_ in enumerate(x_history)], dim=-1).sum(dim=-1)
                x = activation(layer(x)+skip_x)
                x_history.append(x)
                dest += 1
            else:
                x = activation(layer(x))

        return x

class CNN(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding), activation_list[0]]
        # Hidden layers ~ Output layer
        for i in range(len(n_channel_list) - 1):
            layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding), activation_list[i+1]])
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)

class CNN_Resnet(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert kernel_size % 2 == 1, f'kernel size must be odd number, Got: ({kernel_size})'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding)]
        # Hidden layers ~ Output layer
        layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding) for i in range(len(n_channel_list)-1)])
        self.layers = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_channel_list = [in_channel]+[n_channel_list[i] for i in range(skip-1, len(n_channel_list), skip)]
        skip_layers = [nn.Conv2d(in_channels=n_skip_channel_list[i], out_channels=n_skip_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding, bias=False) for i in range(len(n_skip_channel_list)-1)] # Bias already present at original layers
        self.skip_layers = nn.ModuleList(skip_layers) # Need ModuleList, not list, so that the model can recognize this layers
        self.n_skip = len(n_channel_list) // skip # number of skip layers
        # assert self.n_skip == len(self.skip_layers), 'something\'s wrong'

    def forward(self, x):
        last_x = x
        skip_layers = iter(self.skip_layers)
        for i, (layer, activation) in enumerate(zip(self.layers, self.activation_list)):
            if (i+1) % self.skip == 0:
                skip_layer = next(skip_layers)
                x = activation(layer(x)+skip_layer(last_x))
                last_x = x
            else:
                x = activation(layer(x))
        return x

class CNN_Densenet(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [dcopy(activation_list)]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert kernel_size % 2 == 1, f'kernel size must be odd number, Got: ({kernel_size})'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding)]
        # Hidden layers ~ Output layer
        layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding) for i in range(len(n_channel_list)-1)])
        self.layers = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_channel_list = [in_channel]+[n_channel_list[i] for i in range(skip-1, len(n_channel_list), skip)]

        skip_layers = nn.ModuleDict()
        for i in range(1, len(n_skip_channel_list)):
            for j in range(i):
                skip_layers[str(j)+'_'+str(i)]=nn.Conv2d(in_channels=n_skip_channel_list[j], out_channels=n_skip_channel_list[i], kernel_size=kernel_size, stride=1, padding=padding, bias=False)
        self.skip_layers = skip_layers

        n = len(n_skip_channel_list)-1
        self.n_skip = int(n*(n+1)/2)
        assert self.n_skip == len(skip_layers), f'something wrong: n_skip({self.n_skip}), skip_layers({len(skip_layers)})'

    def forward(self, x):
        x_history = [x]
        dest = 1 # Destination index of skip layers
        for i, (layer, activation) in enumerate(zip(self.layers, self.activation_list)):
            if (i+1) % self.skip == 0:
                assert dest == len(x_history), f'Length of x_history{len(x_history)} and destination number{dest} must match'
                skip_x = torch.stack([self.skip_layers[str(src)+'_'+str(dest)](x_) for src, x_ in enumerate(x_history)], dim=-1).sum(dim=-1)
                x = activation(layer(x)+skip_x)
                x_history.append(x)
                dest += 1
            else:
                x = activation(layer(x))

        return x

class Gaussian(nn.Module):
    def __init__(self, amp=1, mu=0, sigma=1):
        super().__init__()
        self.register_parameter('mu', nn.Parameter(torch.tensor(mu, dtype=torch.float)))
        self.register_parameter('sigma', nn.Parameter(torch.tensor(sigma, dtype=torch.float)))
        self.register_parameter('amp', nn.Parameter(torch.tensor(amp, dtype=torch.float)))

    def forward(self, x):
        return self.amp/(self.sigma*np.sqrt(2*np.pi))*torch.exp(-0.5*((x-self.mu)/self.sigma)**2)

class SimpleGaussian(nn.Module):
    def __init__(self, amp=1, mu=0, sigma=1):
        super().__init__()
        self.register_parameter('mu', nn.Parameter(torch.tensor(mu, dtype=torch.float)))
        self.register_parameter('sigma', nn.Parameter(torch.tensor(sigma, dtype=torch.float)))
        self.register_parameter('amp', nn.Parameter(torch.tensor(amp, dtype=torch.float)))

    def forward(self, x):
        return self.amp*torch.exp(-0.5*((x-self.mu)/self.sigma)**2)

class Residual(nn.Module):
    """
    Residual module which computes residual connections in forward pass.
    Note that intermediate tensors need to have the same shape.

    Parameters
    ----------
    blocks : list of nn.Module objects, or an iterable nn.Module object.

    activation : nn.Module object (default: None)
        activation function to be applied at each residual connection betwee blocks.

    Examples
    --------
    >>> blocks = [nn.Linear(10,10) for i in range(5)]
    >>> model = Residual(blocks)
    >>> model

    """
    def __init__(self, blocks, activation=None):
        super().__init__()
        if issubclass(type(blocks), nn.Module):
            self.blocks = blocks
        elif issubclass(type(blocks), list):
            self.blocks = nn.ModuleList(blocks)
        else:
            raise Exception('')
        self.activation = activation

    def forward(self, x):
        if self.activation is not None:
            for block in self.blocks:
                x = self.activation(block(x)+x)

        else: # No activation
            for block in self.blocks:
                x = block(x)+x
        return x

class Parallel(nn.Module):
    """
    Similar to nn.Sequential, but parallel connection.
    Note that the shape of each output from each module must be identical,
    since Parallel() sums up all outputs in forward().
    Also note that Parallel with nn.Identity() is identical to Residual()

    Parameters
    ----------
    args : nn.Modules

    Examples
    --------
    >>>
    """
    def __init__(self, *args):
        super().__init__()
        self._layers = args
        for i, layer in enumerate(args):
            self.add_module(str(i), layer)

    def forward(self, x):
        y_list = [layer(x) for layer in self._layers]
        y = sum(y_list)
        return y

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i < len(self):
            layer = self[self._i]
            self._i +=1
            return layer
        else:
            raise StopIteration

    def __len__(self):
        return len(self._layers)

    def __getitem__(self, idx):
        return self._layers[idx]

if __name__ == '__main__':
    n_input = 5
    n_hidden_list = [10,20,30,20,10]
    activation_list = [nn.Sigmoid(), nn.ReLU(), nn.LeakyReLU(0.03), nn.Tanh(), nn.Identity()]
    fnn = FNN(n_input, n_hidden_list, activation_list)
    fnn_resnet = FNN_Resnet(n_input, n_hidden_list, activation_list)
    fnn_densenet = FNN_Densenet(n_input, n_hidden_list, activation_list)
    print(fnn)
    print(fnn_resnet)
    print(fnn_densenet)



import numbers

import numpy as np
import pandas as pd

def isclose(a, b, rtol=1e-09, atol=0.0, equal_nan=False):
    """
    Wraparound np.isclose for pandas.DataFrame

    Parameters
    ----------
    a, b : pandas.DataFrame
        Input arrays to compare.
    rtol : float
        The relative tolerance parameter (see Notes).
    atol : float
        The absolute tolerance parameter (see Notes).
    equal_nan : bool
        Whether to compare NaN's as equal.  If True, NaN's in `a` will be
        considered equal to NaN's in `b` in the output array.

    Returns
    -------
    C : pandas.DataFrame 
        Returns a boolean array of where `a` and `b` are equal within the
        given tolerance. If both `a` and `b` are scalars, returns a single
        boolean value.
    """
    assert a.shape == b.shape, "Shape mismatch"
    assert (a.columns == b.columns).all(), "Column mismatch"
    assert (a.index == b.index).all(), "Index mismatch"

    close_list = []
    for col in a.columns:
        if issubclass(a[col].dtype.type, numbers.Number):
            close = np.isclose(a[col], b[col], rtol=rtol, atol=atol, equal_nan=equal_nan)
        else:
            close = (a[col] == b[col]).to_numpy()
        close_list.append(close)
    close = np.stack(close_list, axis=1)

    return pd.DataFrame(close, columns=a.columns, index=a.index)

if __name__ == '__main__':
    pass



import numpy as np
import scipy.stats as st

def ci(std, n, p=0.95):
    z = st.norm.ppf((1+p)/2) # Two-tailed confidence interval
    half_range = z*std/np.sqrt(n)
    return half_range

def interval(distrib, confidence, **kwargs):
    if 'std' in kwargs and 'n' in kwargs:
        scale = kwargs['std']/np.sqrt(kwargs['n'])
        assert 'scale' not in kwargs, 'scale is given but std and n are also given'
        kwargs['scale'] = scale
        del kwargs['std'], kwargs['n']

    return distrib.interval(confidence, **kwargs)

def ci_stats(mean, std, n, p=0.95):
    '''
    calculate confidence interval based on given statistics
    :param mean: 1d array, sample mean
    :param std: 1d array, sample mean
    :param n: 1d array, sample size
    :param p: 1d array, confidence level
    '''
    # z = st.norm.ppf(p)
    # half_range = z*std/n
    half_range = ci(std=std, n=n)

    confidence_interval = {
    'low': mean-half_range,
    'high': mean+half_range
    }
    return confidence_interval

def ci_samples(x, p=0.95):
    '''
    calculate confidence interval based on samples
    '''
    mean = np.mean(x)
    std = np.std(x)
    n = len(x)
    return ci_stats(mean, std, n, p=p)



from copy import deepcopy as dcopy
import itertools as it
import os
import warnings
import pathlib
import pickle
import subprocess
import time

from . import os as _os

__all__ = \
[
#  'Filename',
'Printer',
'TDict',
'Timer',
 'Path',
'Wrapper',
'append',
'cmd',
'equal',
'equal_set',
'equal_array',
'iseven',
'isodd',
'is_iterable',
'isnumeric',
'isint',
'load_pickle',
'now',
'reverse_dict',
'unnest_dict',
'nestdict_to_list',
'update_ld',
'update_keys',
'merge_dict',
'merge_tuple',
'prettify_dict',
'read',
'readline',
'readlines',
'save_pickle',
'str2bool',
'write'
]

 # def save_pickle(obj, path = None, protocol = None): # Typings
def save_pickle(obj: str, path: str = None, protocol: int = None):
    '''Save object as Pickle file to designated path.
    If path is not given, default to "YearMonthDay_HourMinuteSecond.p" '''
    if path == None:
        path = time.strftime('%y%m%d_%H%M%S.p')
        warnings.warn(f'Be sure to specify specify argument "path"!, saving as {path}...')
    with open(path, 'wb') as f:
        pickle.dump(obj, f, protocol=protocol)

def load_pickle(path: str):
    '''Load Pickle file from designated path'''
    with open(path, 'rb') as f:
        return pickle.load(f)

def write(content: str, path: str, encoding: str = None):
    with open(path, 'w', encoding = encoding) as f:
        f.write(content)

def append(content: str, path: str , encoding: str = None):
    with open(path, 'a', encoding = encoding) as f:
        f.write(content)

def read(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.read()
    return text

def readline(path, encoding = None):
    '''Create generator object which iterates over the file'''
    with open(path, 'r', encoding = encoding) as f:
        for line in f:
            yield line

def readlines(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.readlines()
    return text

def cmd(command: str, shell: bool = False, encoding: str = None):
    '''Run shell command and return stdout'''
    pipe = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding=encoding, shell=shell, text=True)
    return pipe.stdout

def equal(lst):
    '''return True if all elements in iterable is equal'''
    lst_inst = iter(lst)
    try:
        val = next(lst_inst)
    except StopIteration:
        return True

    for v in lst_inst:
        if v!=val:
            return False
    return True

def equal_set(*args):
    '''return True if each element in args has the same set content'''
    return equal([set(elem) for elem in args])

def equal_array(lst):
    '''return True if all elements in array are equal'''
    lst_inst = iter(lst)
    try:
        val = next(lst_inst)
    except StopIteration:
        return True

    for arr in lst_inst:
        if (arr!=val).any():
            return False
    return True

def iseven(i):
    return i%2==0

def isodd(i):
    return i%2==1

def reverse_dict(d):
    '''
    Reverse key:value pair of dict into value:key.
    Thus, values in dict must be hashable
    '''
    d_ = {}
    for k, v in d.items():
        d_[v]=k
    return d_

def unnest_dict(d, format='.'):
    '''
    Unnest nested dictionary.
    Nested keys are concatenated with "format"(defaults with ".") notation.

    ex)
    >>> rating = {'witcher': {'bombs': 5, 'swords': 7}, 'gta': {'guns': 5}, 'lol': {'magic': 5, 'skills': 3}}
    >>> unnest_dict(rating, format='.')
    {
    'witcher.bombs': 5,
    'witcher.swords': 7,
    'gta.guns': 5,
    'lol.magic': 5,
    'lol.skills': 3
    }
    '''
    un_d = {}
    # d = d.copy()
    for k, v in d.items():
        if type(v)==dict:
            un_d_ = unnest_dict(v, format=format)
            for k_, v_ in un_d_.items():
                un_d[str(k)+format+str(k_)] = v_
        else:
            un_d[k] = v
    return un_d

def nestdict_to_list(d, key=None):
    '''
    open 2nd order nested dict to list of dicts.
    If key is given, then key of the 1st order dict will be merged to the dicts as "key: 1st_key"
    If key is None, then the key of 1st order dict will be discarded.
    '''
    l = []
    for k, v in d.items():
        assert type(v) == dict
        d_ = {}
        if key != None:
            d_[key] = k
        d_.update(v)
        l.append(d_)

    return l

def update_ld(ld, d):
    '''add (key, value) to list of dicts'''
    ld = dcopy(ld)
    for d_ in ld:
        d_.update(d)

    return ld

def update_keys(d, d_key, copy=True): # inplace? copy?
    if copy:
        d = dcopy(d)
    for k_old, k_new in d_key.items():
        d[k_new] = d.pop(k_old)
    return d

def merge_dict(ld):
    '''merge list of dicts
    into dict of lists
    ld: list of dicts'''

    # keys = sorted(list(set([d.keys() for d in ld])))
    keys = sorted(list(set(it.chain(*[d.keys() for d in ld]))))
    d_merged = {key:[] for key in keys}
    for d in ld:
        for k, v in d.items():
            d_merged[k].append(v)
    return d_merged

def merge_tuple(lt):
    '''merge list of tuples
    into tuple of lists
    lt: list of tuples'''
    n_tuple = len(lt[0])
    l_merged = [[] for _ in range(n_tuple)]
    for t in lt:
        for i, v in enumerate(t):
            l_merged[i].append(v)
    return tuple(l_merged)

# def prettify_dict(d, print_type='yaml', **kwargs):
#     '''prettify long dictionary
#     Example
#     -------
#     >>> d = {'a':1, 'b':2, 'c':3}
#     >>> print_dict(d)
#
#     '''
#     if print_type == 'yaml':
#         return yaml.dump(d, **kwargs)
#     elif print_type=='pprint':
#         pp = pprint.PrettyPrinter()
#         return pp.pformat(info_basic)
#     else:
#         pass
def prettify_dict(dictionary, indent=0):
    return '\n'.join([' '*indent + str(k) +': '+str(v) if type(v)!=dict else str(k)+':\n'+prettify_dict(v, indent=indent+2) for k, v in dictionary.items()])

# def pipeline(functions, args):
#     from functools import reduce
#     return reduce(lambda acc, func: func(acc), functions, args)

# Used in Argparse
def str2bool(x):
    try:
        return bool(x)
    except:
        pass

    true_list = ['t', 'true', 'y', 'yes', '1']
    false_list = ['f', 'false', 'n', 'no', '0']
    if x.lower() in true_list:
        return True
    elif x.lower() in false_list:
        return False
    else:
        raise Exception('input has to be in one of two forms:\nTrue: %s\nFalse: %s'%(true_list, false_list))

# DEPRECATED: use str.isnumeric()
# def strisfloat(x):
#     try:
#         x=float(x)
#         return True
#     except ValueError:
#         return False

def isnumeric(x):
    try:
        float(x)
        return True
    except ValueError or TypeError:
        return False

def isint(x):
    try:
        int(x)
        return True
    except ValueError or TypeError:
        return False

def is_iterable(x):
    return hasattr(x, '__iter__')

def now(format: str ='-'):
    if format=='-':
        return time.strftime('%Y-%m-%d_%H-%M-%S')
    elif format=='_':
        return time.strftime('%y%m%d_%H%M%S')
    else:
        raise Exception("format has to be one of ['-', '_']")

# DEPRECATED: use pathlib.Path() instead
# class Filename():
#     '''
#     Class to handle Filename with suffix.
#     To call filename with suffix, call the instance.

#     Example
#     -------
#     >>> file = Filename('model', '.pt')
#     >>> file
#     (Filename, name: "model", suffix: ".pt")
#     >>> str(file)
#     "model"
#     >>> file()
#     "model.pt"
#     >>> (file+'1')
#     "model1"
#     >>> (file+'1')()
#     "model1.pt"
#     '''
#     def __init__(self,obj=None, suffix=''):
#         self.name = str(obj)
#         self.suffix = suffix

#     def __call__(self):
#         '''Add suffix and return str'''
#         return self.name+self.suffix

#     def __radd__(self, other):
#         return Filename(other+self.name,suffix=self.suffix)

#     def __add__(self, other):
#         return Filename(self.name+other,suffix=self.suffix)

#     def __mul__(self, other):
#         return Filename(self.name*other,suffix=self.suffix)

#     def __repr__(self):
#         return '(Filename, name: "%s", suffix: "%s")'%(self.name, self.suffix)

#     def __str__(self):
#         return self.name

# DEPRECATED: use pathlib.Path() instead

class Path(pathlib.Path):
    '''
    Joins paths by . syntax
    (Want to use pathlib.Path internally, but currently inherit from str)

    Parameters
    ----------
    path: str (default: '.')
        Notes the default path. Leave for default blank value which means the current working directory.
        So YOU MUST NOT USE "path" AS ATTRIBUTE NAME, WHICH WILL MESS UP EVERYTHING

    Example
    -------
    >>> path = Path('C:/exp')
    >>> path
    path: C:/exp

    >>> path.DATA = 'CelebA'
    >>> path
    path: C:/exp
    DATA: C:/exp/CelebA

    >>> path.PROCESSED = 'processed'
    >>> path.PROCESSED.M1 = 'method1'
    >>> path.PROCESSED.M2 = 'method2'
    >>> path
    path: C:/exp
    DATA: C:/exp/CelebA
    PROCESSED: C:/exp/processed

    >>> path.PROCESSED
    M1: C:/exp/processed/method1
    M2: C:/exp/processed/method2
    -------

    '''
    def __new__(cls, *args):
        if cls is Path:
            cls = WindowsPath2 if os.name == 'nt' else PosixPath2
        self = super().__new__(cls, *args)

        # self._path=Path(*args)
        # self = object.__new__(cls)
        return self
    
    # def __init__(self, *args):

    def __repr__(self):
        return f'tools.Path({super().__str__()})'

    def summary(self, indent=0):
        '''Print out current path, and children'''
        for name, _path in self.__dict__.items():
            print(' '*indent+name+': '+str(_path))
            if issubclass(type(_path), Path):
                _path.summary(indent+2)

            # if name != 'path':
        # print('\n'.join([key+': '+str(value) for key, value in self.__dict__.items()]))

    def __fspath__(self):
        return super().__fspath__()
        # return self._path.__fspath__()

    def __str__(self):
        return super().__str__()
        # return str(self._path)

    def __setattr__(self, key, value):
        # super(Path, self).__setattr__(key, self / value) # self.joinpath(value)
        if key.startswith('_'):
            super(Path, self).__setattr__(key, value)
        elif hasattr(self, key) and hasattr(getattr(self, key), '__call__'):
            raise AttributeError(f'Attribute "{key}" already exists')
        else:
            super(Path, self).__setattr__(key, Path(self / value))
        # if hasattr(self, 'path'):
        #     assert key != 'path', '"path" is a predefined attribute and must not be used. Use some other attribute name'
        #     super(Path, self).__setattr__(key, Path(os.path.join(self._path, value)))
        # else:
        #     super(Path, self).__setattr__(key, value)

    def join(self, *args):
        return Path(os.path.join(self, *args))

    def makedirs(self, exist_ok=True):
        '''Make directories of all children paths
        Be sure to define all folders first, makedirs(), and then define files in Path(),
        since defining files before makedirs() will lead to creating directories with names of files.
        It is possible to ignore paths with "." as all files do, but there are hidden directories that
        start with "." which makes things complicated. Thus, defining folders -> makedirs() -> define files
        is recommended.'''
        for directory in self.__dict__.values():
            if directory != '':
                os.makedirs(str(directory), exist_ok=exist_ok)
                if type(directory) == Path:
                    directory.makedirs(exist_ok=exist_ok)

    # DEPRECATED: Use shutil.rmtree(path) instead
    # def clear(self, ignore_errors=True):
    #     '''Delete all files and directories in current directory'''
    #     for directory in self.__dict__.values():
    #         shutil.rmtree(directory, ignore_errors=ignore_errors)

    def listdir(self, join=False, isdir=False, isfile=False):
        return _os.listdir(self, join=join, isdir=isdir, isfile=isfile)

class WindowsPath2(Path, pathlib.WindowsPath):
    pass
class PosixPath2(Path, pathlib.PosixPath):
    pass


class TDict(dict):
    '''
    Dictionary which can get items via attribute notation (class.attribute)

    Parameters
    ----------
    Identical to dict()

    Example
    -------

    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __delattr__(self, key):
        self.__delitem__(key)

class Printer():
    def __init__(self, filewrite_dir = None):
        self.content = ''
        self.filewrite_dir = filewrite_dir

    def add(self, text):
        self.content += text

    def print(self, *args, end='\n', flush=False):
        self.add(' '.join([str(arg) for arg in args]))
        print(self.content, end=end, flush=flush)
        if self.filewrite_dir != None:
            append(self.content + end, self.filewrite_dir)
        self.content=''

    def reset(self):
        self.content = ''

class Timer():
    '''
    Timer to measure elapsed time

    Parameters
    ----------
    print: bool (default: False)
        if True, then prints self.elapsed_time whenever stop() is called.

    return_f: bool (default: True)
        if True, then returns self.elapsed_time whenever stop() is called.

    auto_reset: bool (default: True)
        if True, then resets whenever start() is called.

    Methods
    -------
    start

    stop

    reset
        sets self.elapsed_time = 0
    '''
    def __init__(self, auto_reset = True):
        self.auto_reset = auto_reset
        self._elapsed_time = 0
        self.running = False

    def __repr__(self):
        options = ''
        if self.auto_reset: options += '(auto_reset)'

        if len(options)==0:
            return f'[Timer][running: {self.running}][elapsed_time: {self.elapsed_time()}]'
        else:
            return f'[Timer {options}][running: {self.running}][elapsed_time: {self.elapsed_time()}]'

    def __enter__(self):
        self.start()

    def __exit__(self):
        self.stop()

    def start(self):
        if self.auto_reset:
            # self.reset()
            self._elapsed_time = 0
        self.running = True
        self.start_time = time.time()

    def stop(self):
        if self.running:
            self.end_time = time.time()
            self._elapsed_time += self.end_time - self.start_time
            self.running = False
        return self._elapsed_time

    def elapsed_time(self):
        if self.running:
            return self._elapsed_time + time.time() - self.start_time
        else:
            return self._elapsed_time

    def reset(self):
        self._elapsed_time = 0

class Wrapper:
    def __repr__(self):
        name = '<wrapper>\n'
        args = 'args: '+' '.join(str(self.args))+'\n'
        kwargs = 'kwargs: '+str(self.kwargs)
        return name+args+kwargs

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = T.TDict(kwargs)




from . import federated_learning, model
from ._pandas import *
from .utils import *
from . import data

def to_onehot(x, max_dim):
    '''
    x: list of int

    max_dim: maximum dimension
    '''
    onehot = torch.zeros(len(x),max_dim)
    onehot[range(len(x)), x] = 1
    return onehot



from ._f import *
from ._utils import *

if __name__ == '__main__':
    ld = [{i:i+1 for i in range(5)} for j in range(3)]
    merge_dict(ld)
    ld = [{i:np.arange(i+1) for i in range(5)} for j in range(3)]
    merge_dict(ld)



import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as D
from sklearn.base import BaseEstimator
import sys
import time
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class NNEstimator(BaseEstimator):
    def __init__(self, model=None, criterion=None, val_dataset=None, epoch=10, lr=0.001, batch_size=32, patience=20, device=device, log_dir=None):
        # All arguments need default values. (sklearn)
        # All arguments must be fed into self.args = args
        self.model = model
        self.criterion = criterion

        self.epoch = epoch
        self.lr = lr
        self.batch_size = batch_size
        self.device = device

        # Optional arguments
        self.val_dataset = val_dataset
        self.log_dir = log_dir

    def fit(self, x, y):
        '''
        :param x: numpy.array
        :param y: nump.array
        '''
        self.model.to(self.device)
        op = optim.Adam(self.model.parameters(), lr=self.lr)

        x = torch.as_tensor(x, torch.float32)
        y = torch.as_tensor(y)

        loader = D.DataLoader(dataset, batch_size=self.batch_size, shuffle=True)
        for epoch in range(self.epoch):
            for x, y in self.loader:
                x, y = x.to(self.device), y.to(self.device)
                y_hat = self.model(x)
                loss = self.criterion(y, y_hat)

                op.zero_grad()
                loss.backward()
                op.step()

                if self.val_dataset is not None:
                    validation
                else:
                    loss.item()

        self.model.cpu()

        # if self.log_dir is None:
        #     log_dir = open(path.EXP.join(time.strftime('%Y-%m-%d_%H-%M-%S')+'.txt'))
        # default_stdout = sys.stdout
        # sys.stdout = log_dir


        # All parameters created here should be fed into self.args_ = args

    def predict(self, x):
        # return predicted value
        return y



import numpy as np

# %%
__all__ = [
'angle',
'binarize',
'ceil',
'floor',
'moving_mean',
'standardize',
]

# %%
def floor(x, decimals=0):
    if decimals==0:
        return np.floor(x)
    else:
        return np.floor(x * 10**decimals) / 10**decimals

def ceil(x, decimals=0):
    if decimals==0:
        return np.ceil(x)
    else:
        return np.ceil(x * 10**decimals) / 10**decimals

def standardize(array, axis=None, ep=1e-20):
    return (array - array.mean(axis=axis))/(array.std(axis=axis)+ep)

def binarize(array, threshold):
    '''binarize array with array>=threshold == 1'''
    result = np.zeros_like(array)
    result[array>=threshold] = 1
    return result

def moving_mean(x, w):
    odd = bool(w%2)
    edge = w//2
    if odd:
        x = np.pad(x, w//2+1, mode='edge')
        x = np.cumsum(x).astype(np.float64)
        x = (x[w:] - x[:-w])/w
        return x[:-1]
    else:
        x = np.pad(x, w//2, mode='edge')
        x = np.cumsum(x).astype(np.float64)
        x = (x[w:] - x[:-w])/w
        return x

def angle(x1, x2):
    '''angle between two vectors, derived from cosine rule
    return theta within range of [0,np.pi]'''
    theta = np.arccos(x1@x2/(np.linalg.norm(x1)*np.linalg.norm(x2)))
    return theta

def sigmoid(x):
    return 1/(1+np.exp(-x))

# %%
if __name__ == '__main__':
    pass



from tools import __version__
print(f'Installing tools@{__version__}')

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tools-jsyoo61", # Replace with your own username
    version=__version__,
    author="JaeSung Yoo",
    author_email="jsyoo61@unc.edu",
    description="personal syntax tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jsyoo61/tools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    # install_requires=['numpy','pandas','matplotlib']
)



import sklearn.metrics as metrics
import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt

__all__ = [
'confusion_matrix',
'plot_confusion_matrix',
]

def plot_confusion_matrix(y_true, y_pred, ax=None, **kwargs):
    '''Just a wrapper to generate the confusion matrix and the plot'''
    c_matrix = metrics.confusion_matrix(y_true, y_pred)
    return confusion_matrix(c_matrix, ax=ax)

def confusion_matrix(c_matrix, ax=None, **kwargs):
    '''Plot confusion matrix heatmap'''
    df_cm = pd.DataFrame(c_matrix)
    if ax is None:
        ax = sns.heatmap(df_cm, annot=True, **kwargs)
    else:
        ax = sns.heatmap(df_cm, annot=True, ax=ax, **kwargs)
    ax.set_ylabel('y_true')
    ax.set_xlabel('y_pred')
    return ax

if __name__ == '__main__':
    pass



import os
from typing import Union
__all__ = \
[
'listdir',
'makedirs'
]

def makedirs(path, exist_ok=True):
    os.makedirs(path, exist_ok=exist_ok)

def listdir(path=None, isdir: bool=False, isfile: Union[bool, str]=False, join: bool=False):
    '''
    :func listdir:

    :param path:
    :param isdir:
    :param isfile:
    to get files which start with ".", set isfile=''.
    :param join:
    '''
    # Safety check
    if type(isfile)==str:
        ext = isfile.lower() # extension
        isfile_str = True
        isfile = True
    else:
        isfile_str=False
    # assert type(isfile)==str or type(isfile)==bool, 'isfile can be either str or bool'
    # _isfile = True if type(isfile)==str else isfile
    assert not (isdir and isfile), 'only one of argument "isdir" and "isfile" can be True'
    # if path == None:
    #     path = '.'

    dir_list = os.listdir(path)

    # when path is referring to somewhere non-cwd, we need dir_list_joined to pass reference for isdir() or isfile().
    if join or isdir or isfile:
        dir_list_joined = dir_list if path is None else [os.path.join(path, dir) for dir in dir_list]

    if join:
        dir_list = dir_list_joined

    if isdir:
        return [dir for dir, dir_joined in zip(dir_list, dir_list_joined) if os.path.isdir(dir_joined)]
    elif isfile:
        if isfile_str:
            return [dir for dir, dir_joined in zip(dir_list, dir_list_joined) if os.path.isfile(dir_joined) and (os.path.splitext(dir)[1].lower() == ext)]
        else:
            return [dir for dir, dir_joined in zip(dir_list, dir_list_joined) if os.path.isfile(dir_joined)]
    else:
        return dir_list

# if __name__ == '__main__':
#     listdir('torch', isdir=False, join=False)
#     listdir('torch', isdir=False, join=True)
#     listdir('torch', isdir=True, join=False)
#     listdir('torch', isdir=True, join=True)
#
#     listdir('torch', isfile=True, join=False)
#     listdir('torch', isfile=True, join=True)
#     listdir('torch', isfile='.p', join=False)
#     listdir('torch', isfile='.py', join=True)
#     listdir('torch', isfile='', join=True)
#
#     listdir('torch', isdir=True, isfile=True)
#     listdir('torch', isdir=True, isfile='py')
